# -*- coding: utf-8 -*-

import os

from mmpy_bot_mk2 import __version__

os.environ["DJANGO_SETTINGS_MODULE"] = "settings"

extensions = ["sphinx.ext.autodoc", "sphinx.ext.todo"]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = "mmpy_bot_mk2"
copyright = "2016, "
version = __version__
release = __version__
exclude_patterns = []
pygments_style = "sphinx"
html_theme = "default"
htmlhelp_basename = "mmpy_bot_mk2_doc"
latex_documents = [
    ("index", "mmpy_bot_mk2.tex", "mmpy_bot_mk2 Documentation", "", "manual"),
]
man_pages = [("index", "mmpy_bot_mk2", "mmpy_bot_mk2 Documentation", ["gotlium"], 1)]
texinfo_documents = [
    (
        "index",
        "mmpy_bot_mk2",
        "Mattermost-bot Documentation",
        "gotlium",
        "mmpy_bot_mk2",
        "One line description of project.",
        "Miscellaneous",
    ),
]



import asyncio
import logging
import sys
from typing import List, Optional, Union

from mmpy_bot_mk2.driver import Driver
from mmpy_bot_mk2.event_handler import EventHandler
from mmpy_bot_mk2.exceptions import LoginFailed
from mmpy_bot_mk2.plugins import (
    ExamplePlugin,
    HelpPlugin,
    Plugin,
    PluginManager,
    WebHookExample,
)
from mmpy_bot_mk2.settings import Settings
from mmpy_bot_mk2.webhook_server import WebHookServer

log = logging.getLogger("mmpy.bot")


class Bot:
    """Base chatbot class.

    Can be either subclassed for custom functionality, or used as-is with custom plugins
    and settings. To start the bot, simply call bot.run().
    """

    def __init__(
        self,
        settings: Optional[Settings] = None,
        plugins: Optional[Union[List[Plugin], PluginManager]] = None,
        enable_logging: bool = True,
    ):
        self._setup_plugin_manager(plugins)

        # Use default settings if none were specified.
        self.settings = settings or Settings()

        self.console = None

        if enable_logging:
            self._register_logger()

        self.driver = Driver(
            {
                "url": self.settings.MATTERMOST_URL,
                "port": self.settings.MATTERMOST_PORT,
                "token": self.settings.BOT_TOKEN,
                "scheme": self.settings.SCHEME,
                "verify": self.settings.SSL_VERIFY,
                "basepath": self.settings.MATTERMOST_API_PATH,
                "keepalive": True,
                "connect_kw_args": {"ping_interval": None},
            }
        )
        try:
            self.driver.login()
        except Exception as e:
            log.error(
                "Failed to login to Mattermost server with url: %s, port: %s, ssl_verify: %s. Error: %s",
                self.settings.MATTERMOST_URL,
                self.settings.MATTERMOST_PORT,
                self.settings.SSL_VERIFY,
                e,
            )
            raise LoginFailed(
                (
                    f"Failed to login to Mattermost server with "
                    f"url: {self.settings.MATTERMOST_URL}, "
                    f"port: {self.settings.MATTERMOST_PORT}, "
                    f"ssl_verify: {self.settings.SSL_VERIFY}. "
                    f"Error: {e}"
                )
            ) from e
        self.plugin_manager.initialize(self.driver, self.settings)
        self.event_handler = EventHandler(
            self.driver, settings=self.settings, plugin_manager=self.plugin_manager
        )
        self.webhook_server = None

        if self.settings.WEBHOOK_HOST_ENABLED:
            self._initialize_webhook_server()

        self.running = False

    def _setup_plugin_manager(self, plugins):
        if plugins is None:
            self.plugin_manager = PluginManager(
                [HelpPlugin(), ExamplePlugin(), WebHookExample()]
            )
        elif isinstance(plugins, PluginManager):
            self.plugin_manager = plugins
        else:
            self.plugin_manager = PluginManager(plugins)

    def _register_logger(self):
        logging.basicConfig(
            **{
                "format": self.settings.LOG_FORMAT,
                "datefmt": self.settings.LOG_DATE_FORMAT,
                "level": logging.DEBUG if self.settings.DEBUG else logging.INFO,
                "filename": self.settings.LOG_FILE,
                "filemode": "w",
            }
        )
        # define and add a Handler which writes log messages to the sys.stdout
        # avoid double logging if no log file is specified
        if self.settings.LOG_FILE is not None:
            self.console = logging.StreamHandler(stream=sys.stdout)
            self.console.setFormatter(
                logging.Formatter(
                    self.settings.LOG_FORMAT, self.settings.LOG_DATE_FORMAT
                )
            )
            logging.getLogger("").addHandler(self.console)

    def _initialize_webhook_server(self):
        self.webhook_server = WebHookServer(
            url=self.settings.WEBHOOK_HOST_URL, port=self.settings.WEBHOOK_HOST_PORT
        )
        self.driver.register_webhook_server(self.webhook_server)
        # Schedule the queue loop to the current event loop so that it starts together
        # with self.init_websocket.
        asyncio.get_event_loop().create_task(
            self.event_handler._check_queue_loop(self.webhook_server.event_queue)
        )

    def run(self):
        log.info(f"Starting bot {self.__class__.__name__}.")
        try:
            self.running = True

            self.driver.threadpool.start()
            # Start a thread to run potential scheduled jobs
            self.driver.threadpool.start_scheduler_thread(
                self.settings.SCHEDULER_PERIOD
            )
            # Start the webhook server on a separate thread if necessary
            if self.settings.WEBHOOK_HOST_ENABLED:
                self.driver.threadpool.start_webhook_server_thread(self.webhook_server)

            # Trigger "start" methods on every plugin
            self.plugin_manager.start()

            # Start listening for events
            self.event_handler.start()

        except KeyboardInterrupt as e:
            raise e

        finally:
            # When either the event handler finishes (if we asked it to stop) or we
            # receive a KeyboardInterrupt, shut down the bot.
            self.stop()

    def stop(self):
        if not self.running:
            return

        log.info("Stopping bot.")

        # Shutdown the running plugins
        self.plugin_manager.stop()

        # Stop the threadpool
        self.driver.threadpool.stop()
        self.running = False



from mmpy_bot_mk2.plugins.base import Plugin, PluginManager
from mmpy_bot_mk2.plugins.example import ExamplePlugin
from mmpy_bot_mk2.plugins.help_example import HelpPlugin
from mmpy_bot_mk2.plugins.webhook_example import WebHookExample

__all__ = ["Plugin", "PluginManager", "HelpPlugin", "ExamplePlugin", "WebHookExample"]



class MMPYException(Exception):
    """Base class for exceptions in this module."""

    pass


class LoginFailed(MMPYException):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class UserNotFound(MMPYException):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



from pathlib import Path

from mmpy_bot_mk2.bot import Bot
from mmpy_bot_mk2.function import (
    MessageFunction,
    WebHookFunction,
    listen_to,
    listen_webhook,
)
from mmpy_bot_mk2.plugins import ExamplePlugin, Plugin, WebHookExample
from mmpy_bot_mk2.scheduler import schedule
from mmpy_bot_mk2.settings import Settings
from mmpy_bot_mk2.wrappers import ActionEvent, Message, WebHookEvent

__version__ = Path(__file__).parent.joinpath("version.txt").read_text().rstrip()

__all__ = [
    "__version__",
    "Bot",
    "MessageFunction",
    "WebHookFunction",
    "listen_to",
    "listen_webhook",
    "ExamplePlugin",
    "Plugin",
    "WebHookExample",
    "schedule",
    "Settings",
    "ActionEvent",
    "Message",
    "WebHookEvent",
]



from __future__ import annotations

from mmpy_bot_mk2.driver import Driver
from mmpy_bot_mk2.function import listen_to
from mmpy_bot_mk2.plugins.base import Plugin, PluginManager
from mmpy_bot_mk2.settings import Settings
from mmpy_bot_mk2.wrappers import Message


def _custom_help_sort(rec):
    return (
        rec.metadata.get("category", ""),  # No categories first
        rec.help_type,
        rec.pattern.lstrip("^[(-"),
    )


def _prepare_function_help_message(h, string):
    cmd = h.metadata.get("human_description", h.pattern)
    direct = "`(+)`" if h.direct else ""
    mention = "`(*)`" if h.mention else ""

    if h.help_type == "webhook":
        string += f"- `{cmd}` {direct} {mention} - (webhook) {h.docheader}\n"
    else:
        if not h.docheader:
            string += f"- `{cmd}` {direct} {mention}\n"
        else:
            string += f"- `{cmd}` {direct} {mention} - {h.docheader}\n"

    return string


class HelpPlugin(Plugin):
    """Provide a `help` command that lists functions provided by all plugins.

    With a few plugins enabled the help text can become quite verbose. For this reason,
    this plugin defaults to sending a private/direct message with the information.

    If you wish to disable this behavior pass `direct_help=False` and the help text will
    be displayed in the channel where it is requested.
    """

    def __init__(
        self,
        direct_help: bool = True,
    ):
        super().__init__()
        self.direct_help: bool = direct_help

    def initialize(
        self,
        driver: Driver,
        plugin_manager: PluginManager,
        settings: Settings,
    ):
        super().initialize(driver, plugin_manager, settings)

        if self.settings.RESPOND_CHANNEL_HELP:
            self.help = listen_to("^!help$")(self.help)

    def get_help_string(self, message: Message) -> str:
        """Renders help information (FunctionInfo objects) into a markdown string.

        Help information is presented in a condensed format, grouped into categories
        """
        help_function_info = sorted(self.get_help(message), key=_custom_help_sort)

        string = "### The following functions have been registered:\n\n"
        string += "###### `(*)` require the use of `@botname`, "
        string += "`(+)` can only be used in direct message\n"
        old_category = None

        for h in help_function_info:
            # If categories are defined, group functions accordingly
            category = h.metadata.get("category")
            if category != old_category:
                old_category = category
                category = "uncategorized" if category is None else category
                string += f"Category `{category}`:\n"

            string = _prepare_function_help_message(h, string)

        return string

    def get_help(self, message: Message):
        """Obtain Help info from PluginManager.

        Override this method if you need to customize which listeners will be included
        in the help.
        """
        return self.plugin_manager.get_help()

    @listen_to("^help$", needs_mention=True, human_description="help")
    async def help(self, message: Message):
        """Shows this help information."""
        self.driver.reply_to(
            message, self.get_help_string(message), direct=self.direct_help
        )



from mmpy_bot_mk2.driver import Driver
from mmpy_bot_mk2.function import listen_to, listen_webhook
from mmpy_bot_mk2.plugins.base import Plugin, PluginManager
from mmpy_bot_mk2.settings import Settings
from mmpy_bot_mk2.wrappers import ActionEvent, Message, WebHookEvent


class WebHookExample(Plugin):
    """Webhook plugin with examples of webhook server functionality."""

    def initialize(
        self, driver: Driver, plugin_manager: PluginManager, settings: Settings
    ):
        super().initialize(driver, plugin_manager, settings)
        self.webhook_host_url = settings.WEBHOOK_HOST_URL
        self.webhook_host_port = settings.WEBHOOK_HOST_PORT
        return self

    @listen_webhook("ping")
    @listen_webhook("pong")
    async def action_listener(self, event: WebHookEvent):
        """Listens to webhooks 'ping' and 'pong', and either updates the originating
        action post or sends a channel message to indicate that the webhook works."""
        if isinstance(event, ActionEvent):
            self.driver.respond_to_web(
                event,
                {
                    "update": {"message": event.context["text"], "props": {}},
                    "ephemeral_text": "You updated the post!",
                },
            )
        else:
            self.driver.create_post(
                event.body["channel_id"], f"Webhook {event.webhook_id} triggered!"
            )

    @listen_to("!button", direct_only=False)
    async def webhook_button(self, message: Message):
        """Creates a button that will trigger a webhook depending on the choice."""
        self.driver.reply_to(
            message,
            "",
            props={
                "attachments": [
                    {
                        "pretext": None,
                        "text": "Take your pick..",
                        "actions": [
                            {
                                "id": "sendPing",
                                "name": "Ping",
                                "integration": {
                                    "url": f"{self.webhook_host_url}:{self.webhook_host_port}/"
                                    "hooks/ping",
                                    "context": {
                                        "text": "The ping webhook works! :tada:",
                                    },
                                },
                            },
                            {
                                "id": "sendPong",
                                "name": "Pong",
                                "integration": {
                                    "url": f"{self.webhook_host_url}:{self.webhook_host_port}/"
                                    "hooks/pong",
                                    "context": {
                                        "text": "The pong webhook works! :tada:",
                                    },
                                },
                            },
                        ],
                    }
                ]
            },
        )



import asyncio


def spaces(num: int):
    """Utility function to easily indent strings."""
    return " " * num


def completed_future():
    """Utility function to create a stub Future object that asyncio can wait for."""
    future = asyncio.Future()
    future.set_result(True)
    return future


def split_docstring(doc):
    """Split docstring into first line (header) and full body."""
    return (doc.split("\n", 1)[0], doc) if doc is not None else ("", "")



import queue
import warnings
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import mattermostautodriver
from aiohttp.client import ClientSession

from mmpy_bot_mk2.threadpool import ThreadPool
from mmpy_bot_mk2.webhook_server import WebHookServer
from mmpy_bot_mk2.wrappers import Message, WebHookEvent


class Driver(mattermostautodriver.Driver):
    user_id: str = ""
    username: str = ""

    def __init__(self, *args, num_threads=10, **kwargs):
        """Wrapper around the mattermostautodriver Driver with some convenience
        functions and attributes.

        Arguments:
        - num_threads: int, number of threads to use for the default worker threadpool.
        """
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPool(num_workers=num_threads)
        # Queue to communicate with the WebHookServer
        self.response_queue: Optional[queue.Queue] = None
        self.webhook_url = None

    def login(self, *args, **kwargs):
        super().login(*args, **kwargs)
        self.user_id = self.client._userid
        self.username = self.client._username

    def register_webhook_server(self, server: WebHookServer):
        self.response_queue = server.response_queue
        self.webhook_url = f"{server.url}:{server.port}/hooks"

    def create_post(
        self,
        channel_id: str,
        message: str,
        file_paths: Optional[Sequence[str]] = None,
        root_id: str = "",
        props: Dict = {},
        ephemeral_user_id: Optional[str] = None,
    ):
        """Create a post in the specified channel with the specified text.

        Supports sending ephemeral messages if bot permissions allow it. If any file
        paths are specified, those files will be uploaded to mattermost first and then
        attached.
        """
        if file_paths is None:
            file_paths = []

        file_ids = (
            self.upload_files(file_paths, channel_id) if len(file_paths) > 0 else []
        )

        post = dict(
            channel_id=channel_id,
            message=message,
            file_ids=file_ids,
            root_id=root_id,
            props=props,
        )

        if ephemeral_user_id:
            return self.posts.create_post_ephemeral(
                {
                    "user_id": ephemeral_user_id,
                    "post": post,
                }
            )

        return self.posts.create_post(post)

    def get_thread(self, post_id: str):
        warnings.warn(
            "get_thread is deprecated. Use get_post_thread instead", DeprecationWarning
        )
        return self.get_post_thread(post_id)

    def get_post_thread(self, post_id: str):
        """Wrapper around driver.posts.get_post_thread, which for some reason returns
        duplicate and wrongly ordered entries in the ordered list."""
        thread_info = self.posts.get_post_thread(post_id)

        id_stamps = []
        for id, post in thread_info["posts"].items():
            id_stamps.append((id, int(post["create_at"])))
        # Sort the posts by their timestamps
        sorted_stamps = sorted(id_stamps, key=lambda x: x[-1])
        # Overwrite the order with the sorted list
        thread_info["order"] = [id for id, stamp in sorted_stamps]
        return thread_info

    def get_user_info(self, user_id: str):
        """Returns a dictionary of user info."""
        return self.users.get_user(user_id)

    def react_to(self, message: Message, emoji_name: str):
        """Adds an emoji reaction to the given message."""
        return self.reactions.save_reaction(
            {
                "user_id": self.user_id,
                "post_id": message.id,
                "emoji_name": emoji_name,
            },
        )

    def reply_to(
        self,
        message: Message,
        response: str,
        file_paths: Optional[Sequence[str]] = None,
        props: Dict = {},
        ephemeral: bool = False,
        direct: bool = False,
    ):
        """Reply to the given message.

        Supports sending ephemeral messages if the bot permissions allow it. If the
        message is part of a thread, the reply will be added to that thread.

        Also supports replying privately by setting direct=True.
        """
        if file_paths is None:
            file_paths = []

        if direct and not message.is_direct_message:
            # NOTE we explicitly don't pass root_id as it would refer to a
            # post in a channel not a private session
            direct_args = dict(
                receiver_id=message.user_id,
                message=response,
                file_paths=file_paths,
                props=props,
                ephemeral_user_id=message.user_id if ephemeral else None,
            )

            return self.direct_message(**direct_args)

        reply_args = dict(
            channel_id=message.channel_id,
            message=response,
            root_id=message.reply_id,
            file_paths=file_paths,
            props=props,
            ephemeral_user_id=message.user_id if ephemeral else None,
        )

        return self.create_post(**reply_args)

    def direct_message(
        self,
        receiver_id: str,
        message: str,
        file_paths: Optional[Sequence[str]] = None,
        root_id: str = "",
        props: Dict = {},
        ephemeral_user_id: Optional[str] = None,
    ):
        # Private/direct messages are sent to a special channel that
        # includes the bot and the recipient
        direct_id = self.channels.create_direct_channel([self.user_id, receiver_id])[
            "id"
        ]

        return self.create_post(
            channel_id=direct_id,
            message=message,
            root_id=root_id,
            file_paths=file_paths,
            props=props,
            ephemeral_user_id=ephemeral_user_id,
        )

    def respond_to_web(self, event: WebHookEvent, response):
        """Send a web response to the given WebHookEvent."""
        self.response_queue.put((event.request_id, response))
        event.responded = True

    async def trigger_own_webhook(self, webhook_id: str, data: Dict):
        """Triggers a a webhook with id webhook_id on the running WebHookServer."""
        if not self.webhook_url:
            raise ValueError("The Driver is not aware of any running webhook server!")

        async with ClientSession() as session:
            return await session.post(
                f"{self.webhook_url}/{webhook_id}",
                json=data,
            )

    def upload_files(
        self, file_paths: Sequence[Union[str, Path]], channel_id: str
    ) -> List[str]:
        """Given a list of file paths and the channel id, uploads the corresponding
        files and returns a list their internal file IDs."""
        file_list = []
        for path in file_paths:
            path = Path(path)
            # Note: 'files' should be a name of an expected attribute in the body
            # but seems to be ignored when simply uploading files to mattermost
            file_list.append(
                (
                    "files",
                    (
                        path.name,
                        Path(path).read_bytes(),
                    ),
                )
            )

        result = self.files.upload_file(
            files=file_list, data={"channel_id": channel_id}
        )
        return list(info["id"] for info in result["file_infos"])



import asyncio
import random
import time
from queue import Empty, Queue
from typing import Optional

from aiohttp import web

from mmpy_bot_mk2.wrappers import ActionEvent, WebHookEvent


class NoResponse:
    """Used to notify the request handler that no web response should be sent."""

    pass


def handle_json_error(func):
    async def handler(instance, request: web.Request):
        try:
            return await func(instance, request)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            return web.json_response({"status": "failed", "reason": str(e)}, status=400)

    return handler


class WebHookServer:
    """A small server that listens to incoming webhooks and forwards them to the bot
    EventHandler in the main thread/process."""

    def __init__(
        self,
        url: str,
        port: int,
        event_queue: Optional[Queue] = None,
        response_queue: Optional[Queue] = None,
    ):
        self.app = web.Application()
        self.app_runner = web.AppRunner(self.app)
        self.url = url
        self.port = port
        self.running = False

        # Create queues if necessary.
        self.event_queue = event_queue or Queue()
        self.response_queue = response_queue or Queue()
        self.response_handlers = {}

        # Register /hooks endpoint
        self.app.add_routes([web.post("/hooks/{webhook_id}", self.process_webhook)])

    async def start(self):
        webhook_host_ip = self.url.replace("http://", "")
        await self.app_runner.setup()
        site = web.TCPSite(self.app_runner, webhook_host_ip, self.port)
        await site.start()
        self.running = True

        # Schedule the response awaiting function to the same loop as the web server
        asyncio.get_event_loop().create_task(self._obtain_responses_loop())

    async def stop(self):
        await self.app_runner.cleanup()
        self.running = False

    async def _obtain_responses_loop(self):
        """Checks the response queue for incoming responses and passes them on to the
        functions awaiting them."""
        while True:
            try:
                request_id, response = self.response_queue.get_nowait()
                print(f"Received response {response} for request {request_id}")
                try:
                    if not self.response_handlers[request_id].cancelled():
                        self.response_handlers[request_id].set_result(response)
                    del self.response_handlers[request_id]
                except KeyError:
                    # If this handler already received a response, we can skip this.
                    pass
            except Empty:
                pass
            await asyncio.sleep(0.0001)

    @handle_json_error
    async def process_webhook(self, request: web.Request):
        data = await request.json()
        webhook_id = request.match_info.get("webhook_id", "")
        if "trigger_id" in data:
            # Use the trigger ID to identify this request
            event = ActionEvent(
                data, request_id=data["trigger_id"], webhook_id=webhook_id
            )
        else:
            # Generate an ID based on the current time and a random number.
            event = WebHookEvent(
                data,
                request_id=f"{time.time()}_{random.randint(0, 10000)}",
                webhook_id=webhook_id,
            )
        self.event_queue.put(event)

        # Register a Future object that will signal us when a response has arrived,
        # and wait for it to complete.
        await_response = asyncio.get_event_loop().create_future()
        self.response_handlers[event.request_id] = await_response
        await await_response

        result = await_response.result()
        if result is NoResponse:
            return web.Response(status=200)

        return web.json_response(result)



import collections
import os
import warnings
from dataclasses import dataclass, field, fields
from typing import Optional, Sequence, Union, get_args, get_origin  # type: ignore


def _get_comma_separated_list(string: str, type=str):
    values = string.split(",")
    # Convert to the specified type if necessary.
    if type is not str:
        values = list([type(value) for value in values])
    return values


def _is_valid_option(_type, valid_types):
    """Test whether the specified type is an option with an argument type we can support
    (int, float, str, bool)"""
    maintype = get_origin(_type)
    if maintype is not Union:
        # Optional is an alias for Union
        return False

    subtypes = get_args(_type)

    # We need to cast the value in the end so we need a single type
    # As such the only Union type that allows this is Union[Any, None]
    if len(subtypes) != 2 or type(None) not in subtypes:
        return False

    # The non-None type must be one of the types we can support
    if not set(valid_types).intersection(subtypes):
        return False

    return True


@dataclass
class Settings:
    """Specifies the settings to be used by a chatbot. To run a chatbot, you should
    either create a custom Settings instance with the appropriate values, or modify them
    through environment variables.

    The order of priority in which settings are obtained is as follows:
    1. Environment variables
    2. Constructor arguments, if specified
    3. The default value set here.
    """

    MATTERMOST_URL: str = "https://chat.com"
    MATTERMOST_PORT: int = 443
    MATTERMOST_API_PATH: str = ""
    BOT_TOKEN: str = "token"
    BOT_TEAM: str = "team_name"
    SSL_VERIFY: bool = True
    WEBHOOK_HOST_ENABLED: bool = False
    WEBHOOK_HOST_URL: str = "http://127.0.0.1"
    WEBHOOK_HOST_PORT: int = 8579
    DEBUG: bool = False
    # Respond to channel message "!help" (without @bot)
    RESPOND_CHANNEL_HELP: bool = False
    LOG_FILE: Optional[str] = None
    LOG_FORMAT: str = "[%(asctime)s][%(name)s][%(levelname)s] %(message)s"
    LOG_DATE_FORMAT: str = "%m/%d/%Y %H:%M:%S"

    IGNORE_USERS: Sequence[str] = field(default_factory=list)
    # How often to check whether any scheduled jobs need to be run, default every second
    SCHEDULER_PERIOD: float = 1.0

    SCHEME: str = field(init=False)  # Will be taken from the URL. Defaults to https.

    def __post_init__(self):
        # Check if any fields need to be overridden by environment variables
        self._check_environment_variables()

        if "://" in self.MATTERMOST_URL:
            self.SCHEME, self.MATTERMOST_URL = self.MATTERMOST_URL.split("://")
        else:
            self.SCHEME = "https"

        api_url = "/api/v4"

        if self.MATTERMOST_API_PATH.endswith(api_url):
            warnings.warn(
                (
                    f"MATTERMOST_API_PATH should no longer include {api_url} "
                    "or be set unless you run mattermost in a subfolder "
                    "(example.com/mattermost/)."
                ),
                DeprecationWarning,
            )
            self.MATTERMOST_API_PATH = self.MATTERMOST_API_PATH[: -len(api_url)]

    def _check_environment_variables(self):
        for f in fields(self):
            if f.name in os.environ:
                self._set_field(f, os.environ[f.name])

    def _set_field(self, f, value: str):
        if f.default_factory is list:
            # Assert that the type spec of this attribute is indeed a sequence
            if not issubclass(get_origin(f.type), collections.abc.Sequence):
                raise TypeError(
                    f"Since attribute {f.name} has `default_factory=list`, it "
                    "should be specified as a Sequence (or a subclass of it). "
                    f"The actual type is {f.type}."
                )
            if len(get_args(f.type)) == 0:
                raise TypeError(
                    f"Attribute {f.name} was specified as a Sequence without "
                    "specifying what kind of objects it contains."
                )
            # Use get_args to find out what kind of sequence it is.
            value = _get_comma_separated_list(value, type=get_args(f.type)[0])
        elif f.type in [int, float, str]:  # type: ignore
            value = f.type(value)
        elif f.type is bool:
            value = f.type(value.lower() in ("yes", "y", "true", "1"))
        elif _is_valid_option(f.type, [int, float, str]):
            subtype = set(get_args(f.type)).intersection([int, float, str]).pop()
            value = subtype(value)
        elif _is_valid_option(f.type, [bool]):
            subtype = set(get_args(f.type)).intersection([bool]).pop()
            value = subtype(value.lower() in ("yes", "y", "true", "1"))
        else:
            raise TypeError(
                f"Attribute {f.name} has type {f.type}, which is not supported."
                "Consider updating Settings.__post_init__ to support converting"
                " environment variables to this type."
            )
        setattr(self, f.name, value)



import asyncio
import json
import logging
import queue
import re

from mmpy_bot_mk2.driver import Driver
from mmpy_bot_mk2.plugins import PluginManager
from mmpy_bot_mk2.settings import Settings
from mmpy_bot_mk2.webhook_server import NoResponse
from mmpy_bot_mk2.wrappers import Message, WebHookEvent

log = logging.getLogger("mmpy.event_handler")


class EventHandler(object):
    def __init__(
        self,
        driver: Driver,
        settings: Settings,
        plugin_manager: PluginManager,
        ignore_own_messages=True,
    ):
        """The EventHandler class takes care of the connection to mattermost and calling
        the appropriate response function to each event."""
        self.driver = driver
        self.settings = settings
        self.ignore_own_messages = ignore_own_messages
        self.plugin_manager = plugin_manager

        self._name_matcher = re.compile(rf"^@?{self.driver.username}[:,]?\s?")

    def start(self):
        # This is blocking, will loop forever
        self.driver.init_websocket(self._handle_event)

    def _should_ignore(self, message: Message):
        # Ignore message from senders specified in settings, and maybe from ourself
        return (
            True
            if message.sender_name.lower()
            in (name.lower() for name in self.settings.IGNORE_USERS)
            else False
        ) or (self.ignore_own_messages and message.sender_name == self.driver.username)

    async def _check_queue_loop(self, webhook_queue: queue.Queue):
        log.info("EventHandlerWebHook queue listener started.")
        while True:
            try:
                event = webhook_queue.get_nowait()
                await self._handle_webhook(event)
            except queue.Empty:
                pass
            await asyncio.sleep(0.001)

    async def _handle_event(self, data):
        post = json.loads(data)
        event_action = post.get("event")
        if event_action == "posted":
            await self._handle_post(post)

    async def _handle_post(self, post):
        # For some reason these are JSON strings, so need to parse them first
        for item in ["post", "mentions"]:
            if post.get("data", {}).get(item):
                post["data"][item] = json.loads(post["data"][item])

        # If the post starts with a mention of this bot, strip off that part.
        post["data"]["post"]["message"] = self._name_matcher.sub(
            "", post["data"]["post"]["message"]
        )
        message = Message(post)
        if self._should_ignore(message):
            return

        # Find all the listeners that match this message, and have their plugins handle
        # the rest.
        tasks = []
        for matcher, functions in self.plugin_manager.message_listeners.items():
            match = matcher.search(message.text)
            if match:
                groups = list([group for group in match.groups() if group != ""])
                for function in functions:
                    # Create an asyncio task to handle this callback
                    tasks.append(
                        asyncio.create_task(
                            function.plugin.call_function(
                                function, message, groups=groups
                            )
                        )
                    )
        # Execute the callbacks in parallel
        asyncio.gather(*tasks)

    async def _handle_webhook(self, event: WebHookEvent):
        # Find all the listeners that match this webhook id, and have their plugins
        # handle the rest.
        tasks = []
        for matcher, functions in self.plugin_manager.webhook_listeners.items():
            match = matcher.search(event.webhook_id)
            if match:
                for function in functions:
                    # Create an asyncio task to handle this callback
                    tasks.append(
                        asyncio.create_task(
                            function.plugin.call_function(function, event)
                        )
                    )
        # If this webhook doesn't correspond to any listeners, signal the WebHookServer
        # to not wait for any response
        if len(tasks) == 0:
            self.driver.respond_to_web(event, NoResponse)
        # If it does, execute the callbacks in parallel
        else:
            asyncio.gather(*tasks)



from __future__ import annotations

import logging
import re
from abc import ABC
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Union

from mmpy_bot_mk2.driver import Driver
from mmpy_bot_mk2.function import Function, MessageFunction, WebHookFunction
from mmpy_bot_mk2.settings import Settings
from mmpy_bot_mk2.utils import split_docstring
from mmpy_bot_mk2.wrappers import EventWrapper

log = logging.getLogger("mmpy.plugin_base")


class Plugin(ABC):
    """A Plugin is a self-contained class that defines what functions should be executed
    given different inputs.

    It will be called by the EventHandler whenever one of its listeners is triggered,
    but execution of the corresponding function is handled by the plugin itself. This
    way, you can implement multithreading or multiprocessing as desired.
    """

    def __init__(self):
        self.driver: Optional[Driver] = None
        self.plugin_manager: Optional[PluginManager] = None
        self.settings: Optional[Settings] = None
        self.docstring = self.__doc__ if self.__doc__ != Plugin.__doc__ else None

    def initialize(
        self,
        driver: Driver,
        plugin_manager: PluginManager,
        settings: Settings,
    ):
        self.driver = driver
        self.plugin_manager = plugin_manager
        self.settings = settings

    def on_start(self):
        """Will be called after initialization.

        Can be overridden on the subclass if desired.
        """
        log.debug(f"Plugin {self.__class__.__name__} started!")
        return self

    def on_stop(self):
        """Will be called when the bot is shut down manually.

        Can be overridden on the subclass if desired.
        """
        log.debug(f"Plugin {self.__class__.__name__} stopped!")
        return self

    async def call_function(
        self,
        function: Function,
        event: EventWrapper,
        groups: Optional[Sequence[str]] = [],
    ):
        if function.is_coroutine:
            await function(event, *groups)  # type:ignore
        else:
            # By default, we use the global threadpool of the driver, but we could use
            # a plugin-specific thread or process pool if we wanted.
            self.driver.threadpool.add_task(function, event, *groups)


@dataclass
class FunctionInfo:
    help_type: str
    location: str
    function: Function
    pattern: str
    docheader: str
    docfull: str
    direct: bool
    mention: bool
    is_click: bool
    metadata: Dict


def get_function_characteristics(function):
    """Returns a tuple describing the function user interface.

    Returns (direct_only, needs_mention, help_type)
    """
    if isinstance(function, MessageFunction):
        return (function.direct_only, function.needs_mention, "message")
    elif isinstance(function, WebHookFunction):
        return (False, False, "webhook")
    else:
        raise NotImplementedError(
            f"Unknown/Unsupported listener type: '{type(function)}'"
        )


def generate_plugin_help(
    listeners: Dict[re.Pattern[Any], List[Union[MessageFunction, WebHookFunction]]],
):
    """Build FunctionInfo objects from plugin and function information.

    Returns one FunctionInfo instance for every listener (message or webhook)
    """

    plug_help: List[FunctionInfo] = []

    for matcher, functions in listeners.items():
        for function in functions:
            plug_head, plug_full = split_docstring(function.plugin.__doc__)
            func_head, func_full = split_docstring(function.docstring)
            direct, mention, help_type = get_function_characteristics(function)

            plug_help.append(
                FunctionInfo(
                    help_type=help_type,
                    location=function.plugin.__class__.__name__,
                    function=function,
                    pattern=matcher.pattern,
                    docheader=func_head,
                    docfull=func_full,
                    direct=direct,
                    mention=mention,
                    is_click=function.is_click_function,
                    metadata=function.metadata,
                )
            )

    return plug_help


class PluginManager:
    """PluginManager is responsible for initializing all plugins and display aggregated
    help from each of them.

    It is supposed to be transparent to EventHandler that interacts directly with each
    individual Plugin.
    """

    def __init__(
        self,
        plugins: Sequence[Plugin],
    ):
        self.settings: Optional[Settings] = None
        self.plugins = plugins

        self.message_listeners: Dict[re.Pattern, List[MessageFunction]] = defaultdict(
            list
        )
        self.webhook_listeners: Dict[re.Pattern, List[WebHookFunction]] = defaultdict(
            list
        )

    def initialize(self, driver: Driver, settings: Settings):
        for plugin in self.plugins:
            plugin.initialize(driver, self, settings)

            # Register listeners for any listener functions in the plugin
            for attribute in dir(plugin):
                attribute = getattr(plugin, attribute)
                if not isinstance(attribute, Function):
                    continue

                # Register this function and any potential siblings
                for function in [attribute] + attribute.siblings:
                    # Plugin message/webhook handlers can be decorated multiple times
                    # resulting in multiple siblings that do not have .plugin defined
                    # or where the relationship with the parent plugin is incorrect
                    function.plugin = plugin
                    if isinstance(function, MessageFunction):
                        self.message_listeners[function.matcher].append(function)
                    elif isinstance(function, WebHookFunction):
                        self.webhook_listeners[function.matcher].append(function)
                    else:
                        raise TypeError(
                            f"{plugin.__class__.__name__} has a function of unsupported"
                            f" type {type(function)}."
                        )

    def start(self):
        """Trigger on_start() on every registered plugin."""
        for plugin in self.plugins:
            plugin.on_start()

    def stop(self):
        """Trigger on_stop() on every registered plugin."""
        for plugin in self.plugins:
            plugin.on_stop()

    def get_help(self) -> List[FunctionInfo]:
        """Returns a list of FunctionInfo items for every registered message and webhook
        listener."""
        plug_help = generate_plugin_help(self.message_listeners)
        plug_help += generate_plugin_help(self.webhook_listeners)

        return plug_help



from functools import cached_property
from typing import Dict


class EventWrapper:
    """Wrapper around the body of a mattermost network event, e.g. new posts or webhook
    requests. Contains cached properties for convenient variable access.

    Arguments:
    - body: dictionary, body of the network request that contains this event.
    """

    def __init__(
        self,
        body: Dict,
    ):
        self.body = body


class Message(EventWrapper):
    @cached_property
    def id(self):
        return self.body["data"]["post"]["id"]

    @cached_property
    def user_id(self):
        return self.body["data"]["post"]["user_id"]

    @cached_property
    def text(self):
        return self.body["data"]["post"]["message"].strip()

    @cached_property
    def channel_id(self):
        return self.body["data"]["post"]["channel_id"]

    @cached_property
    def channel_name(self):
        return self.body["data"]["channel_name"]

    @cached_property
    def is_direct_message(self):
        return self.body["data"]["channel_type"] == "D"

    @cached_property
    def mentions(self):
        return self.body["data"].get("mentions", [])

    @cached_property
    def parent_id(self):
        return self.body["data"]["post"]["parent_id"]

    @cached_property
    def reply_id(self):
        return self.root_id or self.id

    @cached_property
    def root_id(self):
        return self.body["data"]["post"]["root_id"]

    @cached_property
    def sender_name(self):
        return self.body["data"].get("sender_name", "").strip().strip("@")

    @cached_property
    def team_id(self):
        return self.body["data"].get("team_id", "").strip()


class WebHookEvent(EventWrapper):
    """Wrapper around an incoming webhook post request.

    Arguments:
    - request_id: str, unique identifier of this web request
    - webhook_id: str, the webhook id that was triggered.
    """

    def __init__(
        self,
        *args,
        request_id: str,
        webhook_id: str,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.request_id = request_id
        self.webhook_id = webhook_id
        # Whether a web response was already sent to this request or not.
        self.responded = False

    @cached_property
    def text(self):
        return self.body.get("text")

    @cached_property
    def channel_name(self):
        return self.body.get("channel", self.body.get("channel_name"))

    @cached_property
    def props(self):
        return self.body.get("props", {})

    @cached_property
    def type(self):
        return self.body.get("type")


class ActionEvent(WebHookEvent):
    """Wrapper around an incoming webhook event that was triggered by an action, e.g.
    pressing a button or submitting a form."""

    @cached_property
    def channel_id(self):
        return self.body.get("channel_id")

    @cached_property
    def context(self):
        return self.body.get("context")

    @cached_property
    def data_source(self):
        return self.body.get("data_source")

    @cached_property
    def post_id(self):
        return self.body.get("post_id")

    @cached_property
    def team_id(self):
        return self.body.get("team_id")

    @cached_property
    def trigger_id(self):
        return self.body.get("trigger_id")

    @cached_property
    def user_id(self):
        return self.body.get("user_id")

    @cached_property
    def user_name(self):
        return self.body.get("user_name")



from __future__ import annotations

import asyncio
import inspect
import logging
import re
import shlex
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Optional, Sequence, Union

import click

from mmpy_bot_mk2.utils import completed_future
from mmpy_bot_mk2.webhook_server import NoResponse
from mmpy_bot_mk2.wrappers import Message, WebHookEvent

if TYPE_CHECKING:
    from mmpy_bot_mk2.plugins import Plugin


log = logging.getLogger("mmpy.function")


class Function(ABC):
    def __init__(
        self,
        function: Union[Function, click.Command],
        matcher: re.Pattern,
        **metadata,
    ):
        # If another Function was passed, keep track of all these siblings.
        # We later use them to register not only the outermost Function, but also any
        # stacked ones.
        self.siblings = []

        while isinstance(function, Function):
            self.siblings.append(function)
            function = function.function

        if function is None:
            raise ValueError(
                "ERROR: Possible bug, inside the Function class function should not end up being None"
            )

        self.function = function
        self.is_coroutine = asyncio.iscoroutinefunction(function)
        self.is_click_function: bool = False
        self.matcher = matcher
        self.metadata = metadata

        if not isinstance(function, click.Command):
            self.function.callback = None
            self.function.get_help = None
            self.function.make_context = None
            self.function.invoke = None

        # To be set in the child class or from the parent plugin
        self.plugin: Optional[Plugin] = None
        self.name: Optional[str] = None
        self.docstring = self.function.__doc__ or ""

        @abstractmethod
        def __call__(self, *args):
            pass


class MessageFunction(Function):
    """Wrapper around a Plugin class method that should respond to certain Mattermost
    messages."""

    def __init__(
        self,
        *args,
        direct_only: bool = False,
        needs_mention: bool = False,
        silence_fail_msg: bool = False,
        allowed_users: Optional[Sequence[str]] = None,
        allowed_channels: Optional[Sequence[str]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.is_click_function = isinstance(self.function, click.Command)
        self.direct_only = direct_only
        self.needs_mention = needs_mention
        self.silence_fail_msg = silence_fail_msg

        if allowed_users is None:
            self.allowed_users = []
        else:
            self.allowed_users = [user.lower() for user in allowed_users]

        if allowed_channels is None:
            self.allowed_channels = []
        else:
            self.allowed_channels = [channel.lower() for channel in allowed_channels]

        # Default for non-click functions
        _function: Union[Callable, click.Command] = self.function

        if self.is_click_function:
            _function = self.function.callback
            if asyncio.iscoroutinefunction(_function):
                raise ValueError(
                    "Combining click functions and coroutines is currently not supported!"
                    " Consider using a regular function, which will be threaded by default."
                )
            with click.Context(
                self.function,
                info_name=self.matcher.pattern.strip("^").split("(?: |$)(.*)?")[0],
            ) as ctx:
                # Get click help string and do some extra formatting
                self.docstring += f"\n\n{self.function.get_help(ctx)}"

        if _function is not None:
            self.name = _function.__qualname__

        argspec = list(inspect.signature(_function).parameters.keys())
        if not argspec[:2] == ["self", "message"]:
            raise TypeError(
                "Any message listener function should at least have the positional"
                f" arguments `self` and `message`, but function {self.name} has"
                f" arguments {argspec}."
            )

    def __call__(self, message: Message, *args):
        # We need to return this so that if this MessageFunction was called with `await`,
        # asyncio doesn't crash.
        return_value = None if not self.is_coroutine else completed_future()

        # Check if this message meets our requirements
        if self.direct_only and not message.is_direct_message:
            return return_value

        if self.needs_mention and not (
            message.is_direct_message or self.plugin.driver.user_id in message.mentions
        ):
            return return_value

        if self.allowed_users and message.sender_name not in self.allowed_users:
            if self.silence_fail_msg is False:
                self.plugin.driver.reply_to(
                    message, "You do not have permission to perform this action!"
                )
            return return_value

        if self.allowed_channels and message.channel_name not in self.allowed_channels:
            if self.silence_fail_msg is False:
                self.plugin.driver.reply_to(
                    message, "You do not have permission to perform this action!"
                )
            return return_value

        if self.is_click_function:
            assert len(args) <= 1  # There is only one group, (.*)?
            if len(args) == 1:
                # Turn space-separated string into list
                args = tuple(shlex.split(args[0]))
            try:
                ctx = self.function.make_context(
                    info_name=self.plugin.__class__.__name__, args=list(args)
                )
                ctx.params.update({"self": self.plugin, "message": message})
                return self.function.invoke(ctx)
            # If there are any missing arguments or the function is otherwise called
            # incorrectly, send the click message back to the user and print help string.
            except (click.exceptions.ClickException, click.exceptions.Exit) as e:
                if isinstance(e, click.exceptions.Exit):
                    e = "Requested `--help`:"

                return self.plugin.driver.reply_to(message, f"{e}\n{self.docstring}")

        return self.function(self.plugin, message, *args)


def listen_to(
    regexp: str,
    regexp_flag: int = 0,
    *,
    direct_only=False,
    needs_mention=False,
    allowed_users=None,
    allowed_channels=None,
    silence_fail_msg=False,
    **metadata,
):
    """Wrap the given function in a MessageFunction class so we can register some
    properties."""

    if allowed_users is None:
        allowed_users = []

    if allowed_channels is None:
        allowed_channels = []

    def wrapped_func(func):
        reg = regexp
        if isinstance(func, click.Command):
            if "$" in regexp:
                raise ValueError(
                    f"Regexp of function {func.callback} contains a $, which is not"
                    " supported! The regexp should simply reflect the argument name, and"
                    " click will take care of the rest."
                )

            # Modify the regexp so that it won't try to match the individual arguments.
            # Click will take care of those. We also manually add the ^ if necessary,
            # so that the commands can't be inserted in the middle of a sentence.
            reg = rf"^{reg.strip('^')}(?: |$)(.*)?"  # noqa

        pattern = re.compile(reg, regexp_flag)
        new_func = MessageFunction(
            func,
            matcher=pattern,
            direct_only=direct_only,
            needs_mention=needs_mention,
            allowed_users=allowed_users,
            allowed_channels=allowed_channels,
            silence_fail_msg=silence_fail_msg,
            **metadata,
        )

        # Preserve docstring
        new_func.__doc__ = func.__doc__
        return new_func

    return wrapped_func


class WebHookFunction(Function):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if isinstance(self.function, click.Command):
            raise TypeError(
                "Webhook functions can't be click commands, since they don't take any"
                " additional arguments!"
            )

        self.name = self.function.__qualname__

        argspec = list(inspect.signature(self.function).parameters.keys())
        if not argspec == ["self", "event"]:
            raise TypeError(
                "A webhook listener function should have exactly two arguments:"
                f" `self` and `event`, but function {self.name} has arguments {argspec}."
            )

    def __call__(self, event: WebHookEvent):
        # Signal the WebHookServer that we won't be sending a response.
        def ensure_response(*args):
            if not event.responded:
                self.plugin.driver.respond_to_web(event, NoResponse)

        # If this is a coroutine, wrap it in a task with ensure_response as callback
        if self.is_coroutine:
            task = asyncio.create_task(self.function(self.plugin, event))
            task.add_done_callback(ensure_response)
            return task

        # If not, simply call both of these functions
        try:
            self.function(self.plugin, event)
        except Exception:
            log.exception("Exception occurred: ")
        finally:
            return ensure_response()


def listen_webhook(
    regexp: str,
    **metadata,
):
    """Wrap the given function in a WebHookFunction class with the specified regexp."""

    def wrapped_func(func):
        pattern = re.compile(regexp)
        new_func = WebHookFunction(
            func,
            matcher=pattern,
            **metadata,
        )

        # Preserve docstring
        new_func.__doc__ = func.__doc__
        return new_func

    return wrapped_func



import asyncio
import re
from datetime import datetime
from pathlib import Path

import click
import mattermostautodriver

from mmpy_bot_mk2.function import listen_to
from mmpy_bot_mk2.plugins.base import Plugin
from mmpy_bot_mk2.scheduler import schedule
from mmpy_bot_mk2.wrappers import Message


class ExamplePlugin(Plugin):
    """Default plugin with examples of bot functionality and usage."""

    @listen_to(
        "^admin$", direct_only=True, allowed_users=["admin", "root"], category="admin"
    )
    async def users_access(self, message: Message):
        """Showcases a function with restricted access."""
        self.driver.reply_to(message, "Access allowed!")

    @listen_to("^offtopic_channel$", allowed_channels=["off-topic"], category="admin")
    async def channels_access(self, message: Message):
        """Showcases a function which can only be used in specific channels."""
        self.driver.reply_to(message, "Access allowed!")

    @listen_to("^busy|jobs$", re.IGNORECASE, needs_mention=True, category="admin")
    async def busy_reply(self, message: Message):
        """Show the number of busy worker threads."""
        busy = self.driver.threadpool.get_busy_workers()
        self.driver.reply_to(
            message,
            f"Number of busy worker threads: {busy}",
        )

    @listen_to("hello_click", needs_mention=True, category="click")
    @click.command(help="An example click command with various arguments.")
    @click.argument("POSITIONAL_ARG", type=str)
    @click.option("--keyword-arg", type=float, default=5.0, help="A keyword arg.")
    @click.option("-f", "--flag", is_flag=True, help="Can be toggled.")
    def hello_click(
        self, message: Message, positional_arg: str, keyword_arg: float, flag: bool
    ):
        """A click function documented via docstring."""
        response = (
            "Received the following arguments:\n"
            f"- positional_arg: {positional_arg}\n"
            f"- keyword_arg: {keyword_arg}\n"
            f"- flag: {flag}\n"
        )
        self.driver.reply_to(message, response)

    @listen_to("^hello_channel$", needs_mention=True)
    async def hello_channel(self, message: Message):
        """Responds with a channel post rather than a reply."""
        self.driver.create_post(channel_id=message.channel_id, message="hello channel!")

    # Needs admin permissions
    @listen_to("^hello_ephemeral$", needs_mention=True)
    async def hello_ephemeral(self, message: Message):
        """Tries to reply with an ephemeral message, if the bot has system admin
        permissions."""
        try:
            self.driver.reply_to(message, "hello sender!", ephemeral=True)
        except mattermostautodriver.exceptions.NotEnoughPermissions:
            self.driver.reply_to(
                message, "I do not have permission to create ephemeral posts!"
            )

    @listen_to("^hello_react$", re.IGNORECASE, needs_mention=True)
    async def hello_react(self, message: Message):
        """Responds by giving a thumbs up reaction."""
        self.driver.react_to(message, "+1")

    @listen_to("^hello_file$", re.IGNORECASE, needs_mention=True)
    async def hello_file(self, message: Message):
        """Responds by uploading a text file."""
        file = Path("/tmp/hello.txt")
        file.write_text("Hello from this file!")
        self.driver.reply_to(message, "Here you go", file_paths=[file])

    @listen_to("^!hello_webhook$", re.IGNORECASE, category="webhook")
    async def hello_webhook(self, message: Message):
        """A webhook that says hello."""
        self.driver.client.call_webhook(
            "eauegoqk4ibxigfybqrsfmt48r",
            options={
                "username": "webhook_test",  # Requires the right webhook permissions
                "channel": "off-topic",
                "text": "Hello?",
                "attachments": [
                    {
                        "fallback": "Fallback text",
                        "title": "Title",
                        "author_name": "Author",
                        "text": "Attachment text here...",
                        "color": "#59afe1",
                    }
                ],
            },
        )

    @listen_to("^!info$")
    async def info(self, message: Message):
        """Responds with the user info of the requesting user."""
        user_email = self.driver.get_user_info(message.user_id)["email"]
        reply = (
            f"TEAM-ID: {message.team_id}\nUSERNAME: {message.sender_name}\n"
            f"EMAIL: {user_email}\nUSER-ID: {message.user_id}\n"
            f"IS-DIRECT: {message.is_direct_message}\nMENTIONS: {message.mentions}\n"
            f"MESSAGE: {message.text}"
        )
        self.driver.reply_to(message, reply)

    @listen_to("^ping$", re.IGNORECASE, needs_mention=True)
    async def ping_reply(self, message: Message):
        """Pong."""
        self.driver.reply_to(message, "pong")

    @listen_to(
        "^reply at (.*)$", re.IGNORECASE, needs_mention=True, category="schedule"
    )
    def schedule_once(self, message: Message, trigger_time: str):
        """Schedules a reply to be sent at the given time.

        Arguments:
        - triger_time (str): Timestamp of format %d-%m-%Y_%H:%M:%S,
            e.g. 20-02-2021_20:22:01. The reply will be sent at that time.
        """
        try:
            time = datetime.strptime(trigger_time, "%d-%m-%Y_%H:%M:%S")
            self.driver.reply_to(message, f"Scheduled message at {trigger_time}!")
            schedule.once(time).do(
                self.driver.reply_to, message, "This is the scheduled message!"
            )
        except ValueError as e:
            self.driver.reply_to(message, str(e))

    @listen_to(
        "^schedule every ([0-9]+)$",
        re.IGNORECASE,
        needs_mention=True,
        category="schedule",
    )
    def schedule_every(self, message: Message, seconds: int):
        """Schedules a reply every x seconds. Use the `cancel jobs` command to stop.

        Arguments:
        - seconds (int): number of seconds between each reply.
        """
        schedule.every(int(seconds)).seconds.do(
            self.driver.reply_to, message, f"Scheduled message every {seconds} seconds!"
        )

    @listen_to("^cancel jobs$", re.IGNORECASE, needs_mention=True, category="schedule")
    def cancel_jobs(self, message: Message):
        """Cancels all scheduled jobs, including recurring and one-time events."""
        schedule.clear()
        self.driver.reply_to(message, "Canceled all jobs.")

    @listen_to("^sleep ([0-9]+)$", needs_mention=True)
    async def sleep_reply(self, message: Message, seconds: str):
        """Sleeps for the specified number of seconds.

        Arguments:
            - seconds: How many seconds to sleep for.
        """
        self.driver.reply_to(message, f"Okay, I will be waiting {seconds} seconds.")
        await asyncio.sleep(int(seconds))
        self.driver.reply_to(message, "Done!")



from datetime import datetime
from multiprocessing import Pipe, Process
from multiprocessing.connection import Connection
from typing import Optional

import schedule
from schedule import default_scheduler


class OneTimeJob(schedule.Job):
    # Override schedule.Job._schedule_next_run to avoid periodic job generation.
    def _schedule_next_run(self):
        pass

    def __repr__(self):
        txt = super().__repr__()
        # One time jobs always read "Every 0 None at ..."
        # due undefined self.interval (0) and self.unit (None)
        return txt.replace("Every 0 None", "Once")

    def set_next_run(self, next_time: datetime):
        if not isinstance(next_time, datetime):
            raise AssertionError("The next_time parameter should be a datetime object.")
        self.at_time = next_time
        self.next_run = next_time

    def run(self):
        super().run()
        return schedule.CancelJob()


def _default_scheduler_once(self, trigger_time: datetime):
    job = OneTimeJob(0, self)
    job.set_next_run(trigger_time)
    return job


def _run_job(self, job):
    """Overrides default_scheduler._run_job to support running the jobs in a separate
    process.

    Either way, this waits for the result in a dedicated thread to prevent blocking the
    event loop.
    """
    # Launch job in a dedicated process and send the result through a pipe.
    if "subprocess" in job.tags:

        def wrapped_run(pipe: Connection):
            result = job.run()
            pipe.send(result)

        pipe, child_pipe = Pipe()
        p = Process(target=wrapped_run, args=(child_pipe,))
        p.start()
        # This still blocks despite running in a subprocess
        result = pipe.recv()
    else:
        # Or simply run the job in this thread
        result = job.run()

    if isinstance(result, schedule.CancelJob) or result is schedule.CancelJob:
        self.cancel_job(job)


def _once(trigger_time: Optional[datetime] = None):
    """Adds support for scheduling one-time jobs to the default_scheduler."""
    if trigger_time is None:
        trigger_time = datetime.now()
    if not isinstance(trigger_time, datetime):
        raise AssertionError("The trigger_time parameter should be a datetime object.")
    return default_scheduler.once(self=default_scheduler, trigger_time=trigger_time)


# Monkey-Patching
default_scheduler.once = _default_scheduler_once
schedule.Scheduler._run_job = _run_job
schedule.once = _once



from pathlib import Path

from setuptools import find_packages, setup

excludes = (
    "*test*",
    "*local_settings*",
)


def requires(filename: str):
    return open(filename, encoding="utf-8").read().splitlines()


setup(
    name="mmpy_bot_mk2",
    # Updated by publish workflow
    version=Path(__file__)
    .parent.joinpath("mmpy_bot_mk2/version.txt")
    .read_text(encoding="utf-8")
    .rstrip(),
    author="Alex Tzonkov",
    author_email="anton.ohontsev@gmail.com",
    license="MIT",
    description="A python based bot for Mattermost with its own webhook server.",
    keywords="chat bot mattermost",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/movax01h/mmpy_bot_mk2",
    python_requires=">=3.8",
    platforms=["Any"],
    packages=find_packages(exclude=excludes),
    install_requires=requires("requirements.txt"),
    extras_require={"dev": requires("dev-requirements.txt")},
    package_data={"mmpy_bot_mk2": ["mmpy_bot_mk2/version.txt"]},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "mmpy_bot_mk2 = mmpy_bot_mk2.cli:main",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


SECRET_KEY = "b2!=c))9ju1q(=hlxmzb1l+2qbdjn+*cn(5(u@qn4c-0!2v+%v"
DEBUG = True

ALLOWED_HOSTS = []

INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
)

MIDDLEWARE_CLASSES = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.auth.middleware.SessionAuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "django.middleware.security.SecurityMiddleware",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# Database
# https://docs.djangoproject.com/en/1.8/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

STATIC_URL = "/static/"



import asyncio
import logging
import threading
import time
from queue import Queue

from mmpy_bot_mk2.scheduler import default_scheduler
from mmpy_bot_mk2.webhook_server import WebHookServer

log = logging.getLogger("mmpy.threadpool")


class ThreadPool(object):
    def __init__(self, num_workers: int):
        """Threadpool class to easily specify a number of worker threads and assign work
        to any of them.

        Arguments:
        - num_workers: int, how many threads to run simultaneously.
        """
        self.num_workers = num_workers
        self.alive = False
        self._queue = Queue()
        self._busy_workers = Queue()
        self._threads = []

    def add_task(self, function, *args):
        self._queue.put((function, args))

    def get_busy_workers(self):
        return self._busy_workers.qsize()

    def start(self):
        self.alive = True
        # Spawn num_workers threads that will wait for work to be added to the queue
        for _ in range(self.num_workers):
            worker = threading.Thread(target=self.handle_work)
            self._threads.append(worker)
            worker.start()

    def stop(self):
        """Signals all threads that they should stop and waits for them to finish."""
        self.alive = False
        # Signal every thread that it's time to stop
        for _ in range(self.num_workers):
            self._queue.put((self._stop_thread, tuple()))
        # Wait for each of them to finish
        log.info("Stopping threadpool, waiting for threads...")
        for thread in self._threads:
            thread.join()
        log.info("Threadpool stopped.")

    def _stop_thread(self):
        """Used to stop individual threads."""
        return

    def handle_work(self):
        while self.alive:
            # Wait for a new task (blocking)
            function, arguments = self._queue.get()
            # Notify the pool that we started working
            self._busy_workers.put(1)
            try:
                function(*arguments)
            except Exception:
                log.exception("Unhandled exception in main loop")
            except BaseException:
                # Can be KeyboardInterrupt, SystemExit, ...
                self.alive = False
            # Notify the pool that we finished working
            self._queue.task_done()
            self._busy_workers.get()

    def start_scheduler_thread(self, trigger_period: float):
        def run_pending():
            log.info("Scheduler thread started.")
            while self.alive:
                time.sleep(trigger_period)
                try:
                    default_scheduler.run_pending()
                except Exception:
                    log.exception("Unhandled exception in main loop")
                except BaseException:
                    # Can be KeyboardInterrupt, SystemExit, ...
                    self.alive = False
            log.info("Scheduler thread stopped.")

        self.add_task(run_pending)

    def start_webhook_server_thread(self, webhook_server: WebHookServer):
        async def start_server():
            log.info("Webhook server thread started.")
            await webhook_server.start()
            while self.alive:
                # We just use this to keep the loop running in a non-blocking way
                await asyncio.sleep(0.001)
            await webhook_server.stop()
            log.info("Webhook server thread stopped.")

        self.add_task(asyncio.run, start_server())
